import logging
import os

from functools import partial

import ipykernel
import jupyter_client.session as session
import param

from bokeh.document.events import MessageSentEvent
from bokeh.document.json import Literal, MessageSent, TypedDict
from bokeh.util.serialization import make_id
from ipykernel.comm import Comm, CommManager
from ipykernel.kernelbase import Kernel
from ipywidgets import Widget
from ipywidgets._version import __protocol_version__
from ipywidgets.widgets.widget import _remove_buffers

# Stop ipywidgets_bokeh from patching the kernel
ipykernel.kernelbase.Kernel._instance = ''

from ipywidgets_bokeh.kernel import (
    BokehKernel, SessionWebsocket, WebsocketStream,
)
from ipywidgets_bokeh.widget import IPyWidget
from tornado.ioloop import IOLoop
from traitlets import Any

from ..config import __version__
from ..util import classproperty
from .state import set_curdoc, state

try:
    # Support for ipywidgets>=8.0.5
    import comm

    from comm.base_comm import BaseComm

    class TempComm(BaseComm):
        def publish_msg(self, *args, **kwargs): pass

    comm.create_comm = lambda *args, **kwargs: TempComm(target_name='panel-temp-comm', primary=False)
except Exception:
    comm = None

def _get_kernel(cls=None, doc=None):
    doc = doc or state.curdoc
    if doc is None:
        return _ORIG_KERNEL
    elif doc in state._ipykernels:
        return state._ipykernels[doc]
    state._ipykernels[doc] = kernel = PanelKernel(document=doc, key=str(id(doc)).encode('utf-8'))
    return kernel

def _get_ipywidgets():
    # Support ipywidgets >=8.0 and <8.0
    try:
        from ipywidgets.widgets.widget import _instances as widgets
    except Exception:
        widgets = Widget.widgets
    return widgets

def _on_widget_constructed(widget, doc=None):
    doc = doc or state.curdoc
    if not doc or getattr(widget, '_document', None) not in (doc, None):
        return
    widget._document = doc
    kernel = _get_kernel(doc=doc)
    if (widget.comm and widget.comm.target_name != 'panel-temp-comm' and
        (not (comm and isinstance(widget.comm, comm.DummyComm)) and
         isinstance(widget.comm.kernel, PanelKernel))):
        return
    wstate, buffer_paths, buffers = _remove_buffers(widget.get_state())
    args = {
        'target_name': 'jupyter.widget',
        'data': {
            'state': wstate,
            'buffer_paths': buffer_paths
        },
        'buffers': buffers,
        'metadata': {
            'version': __protocol_version__
        },
        'kernel': kernel
    }
    if widget._model_id is not None:
        args['comm_id'] = widget._model_id
    try:
        widget.comm = Comm(**args)
    except Exception as e:
        if 'PANEL_IPYWIDGET' not in os.environ:
            raise e
    kernel.register_widget(widget)


# Patch font-awesome CSS onto ipywidgets_bokeh IPyWidget
IPyWidget.__css__ = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.5.0/css/font-awesome.css"
]

class MessageSentBuffers(TypedDict):
    kind: Literal["MessageSent"]
    msg_type: str


class MessageSentEventPatched(MessageSentEvent):
    """
    Patches MessageSentEvent with fix that ensures MessageSent event
    does not define msg_data (which is an assumption in BokehJS
    Document.apply_json_patch.)
    """

    def generate(self, references, buffers):
        if not isinstance(self.msg_data, bytes):
            msg = MessageSent(
                kind=self.kind,
                msg_type=self.msg_type,
                msg_data=self.msg_data
            )
        else:
            msg = MessageSentBuffers(
                kind=self.kind,
                msg_type=self.msg_type
            )
            assert buffers is not None
            buffer_id = make_id()
            buf = (dict(id=buffer_id), self.msg_data)
            buffers.append(buf)
        return msg


class PanelSessionWebsocket(SessionWebsocket):

    def __init__(self, *args, **kwargs):
        session.Session.__init__(self, *args, **kwargs)
        self._document = kwargs.pop('document', None)
        self._queue = []
        self._document.on_message("ipywidgets_bokeh", self.receive)

    def send(self, stream, msg_type, content=None, parent=None, ident=None, buffers=None, track=False, header=None, metadata=None):
        msg = self.msg(msg_type, content=content, parent=parent, header=header, metadata=metadata)
        try:
            msg['channel'] = stream.channel
        except Exception:
            return

        packed = self.pack(msg)

        if buffers is not None and len(buffers) != 0:
            buffers = [packed] + buffers
            nbufs = len(buffers)

            start = 4*(1 + nbufs)
            offsets = [start]

            for buffer in buffers[:-1]:
                start += memoryview(buffer).nbytes
                offsets.append(start)

            u32 = lambda n: n.to_bytes(4, "big")
            items = [u32(nbufs)] + [ u32(offset) for offset in offsets ] + buffers
            data = b"".join(items)
        else:
            data = packed.decode("utf-8")

        event = MessageSentEventPatched(self._document, "ipywidgets_bokeh", data)
        self._queue.append(event)
        self._document.add_next_tick_callback(self._dispatch)

    def _dispatch(self):
        try:
            for event in self._queue:
                self._document.callbacks.trigger_on_change(event)
        except Exception as e:
            param.main.param.warning(f'ipywidgets event dispatch failed with: {e}')
        finally:
            self._queue = []

class ShellStream:

    def flush(self, *args):
        pass

class PanelKernel(Kernel):
    implementation = 'panel'
    implementation_version = __version__
    banner = 'banner'

    shell_stream = Any(ShellStream(), allow_none=True)

    def __init__(self, key=None, document=None):
        super().__init__()

        self.session = PanelSessionWebsocket(document=document, parent=self, key=key)
        self.stream = self.iopub_socket = WebsocketStream(self.session)
        self.io_loop = IOLoop.current()

        self.iopub_socket.channel = 'iopub'
        self.session.stream = self.iopub_socket
        self.comm_manager = CommManager(parent=self, kernel=self)
        self.shell = None
        self.session.auth = None
        self.log = logging.getLogger('fake')

        comm_msg_types = ['comm_open', 'comm_msg', 'comm_close']
        for msg_type in comm_msg_types:
            handler = getattr(self.comm_manager, msg_type)
            self.shell_handlers[msg_type] = self._wrap_handler(msg_type, handler)

    async def _flush_control_queue(self):
        pass

    def register_widget(self, widget):
        comm = widget.comm
        comm.kernel = self
        self.comm_manager.register_comm(comm)

    def _wrap_handler(self, msg_type, handler):
        doc = self.session._document
        def wrapper(*args, **kwargs):
            if msg_type == 'comm_open':
                return
            with set_curdoc(doc):
                state.execute(partial(handler, *args, **kwargs), schedule=True)
        return wrapper

# Patch kernel and widget objects
_ORIG_KERNEL = ipykernel.kernelbase.Kernel._instance
if isinstance(ipykernel.kernelbase.Kernel._instance, (BokehKernel, str)):
    ipykernel.kernelbase.Kernel._instance = classproperty(_get_kernel)
Widget.on_widget_constructed(_on_widget_constructed)



from __future__ import annotations

from typing import (
    TYPE_CHECKING, Callable, ClassVar, List, Mapping,
)

import param

from bokeh.models import Column as BkColumn, CustomJS

from ..reactive import Reactive
from .base import NamedListPanel
from .card import Card

if TYPE_CHECKING:
    from bokeh.model import Model

    from ..viewable import Viewable


class Accordion(NamedListPanel):
    """
    The `Accordion` layout is a type of `Card` layout that allows switching
    between multiple objects by clicking on the corresponding card header.

    The labels for each card will default to the `name` parameter of the card’s
    contents, but may also be defined explicitly as part of a tuple.

    Like `Column` and `Row`, `Accordion` has a list-like API that allows
    interactively updating and modifying the cards using the methods `append`,
    `extend`, `clear`, `insert`, `pop`, `remove` and `__setitem__`.

    Reference: https://panel.holoviz.org/reference/layouts/Accordion.html

    :Example:

    >>> pn.Accordion(some_pane_with_a_name, ("Plot", some_plot))
    """

    active_header_background = param.String(default='#ddd', doc="""
        Color for currently active headers.""")

    active = param.List(default=[], doc="""
        List of indexes of active cards.""")

    header_color = param.String(doc="""
        A valid CSS color to apply to the expand button.""")

    header_background = param.String(doc="""
        A valid CSS color for the header background.""")

    toggle = param.Boolean(default=False, doc="""
        Whether to toggle between active cards or allow multiple cards""")

    _bokeh_model = BkColumn

    _direction: ClassVar[str | None] = 'vertical'

    _rename: ClassVar[Mapping[str, str | None]] = {
        'active': None, 'active_header_background': None,
        'header_background': None, 'objects': 'children',
        'dynamic': None, 'toggle': None, 'header_color': None
    }

    _toggle = """
    for (var child of accordion.children) {
      if ((child.id !== cb_obj.id) && (child.collapsed == cb_obj.collapsed) && !cb_obj.collapsed) {
        child.collapsed = !cb_obj.collapsed
      }
    }
    """

    _synced_properties = [
        'active_header_background', 'header_background', 'width',
        'sizing_mode', 'width_policy', 'height_policy', 'header_color',
        'min_width', 'max_width'
    ]

    def __init__(self, *objects, **params):
        super().__init__(*objects, **params)
        self._panels = {}
        self._updating_active = False
        self.param.watch(self._update_active, ['active'])
        self.param.watch(self._update_cards, self._synced_properties)

    def _process_property_change(self, props):
        props.pop('children', None)
        return super()._process_property_change(props)

    def _get_objects(self, model, old_objects, doc, root, comm=None):
        """
        Returns new child models for the layout while reusing unchanged
        models and cleaning up any dropped objects.
        """
        from panel.pane.base import RerenderError, panel
        new_models, old_models = [], []
        if len(self._names) != len(self):
            raise ValueError(
                'Accordion names do not match objects, ensure that the '
                'Accordion.objects are not modified directly. Found '
                f'{len(self._names)} names, expected {len(self)}.'
            )
        for i, (name, pane) in enumerate(zip(self._names, self)):
            pane = panel(pane, name=name)
            self.objects[i] = pane

        for obj in old_objects:
            if obj not in self.objects and id(obj) in self._panels:
                self._panels[id(obj)]._cleanup(root)
                del self._panels[id(obj)]

        params = {
            k: v for k, v in self.param.values().items()
            if k in self._synced_properties
        }

        ref = root.ref['id']
        current_objects = list(self)
        self._updating_active = True
        for i, (name, pane) in enumerate(zip(self._names, self)):
            child_params = dict(params, title=name)
            child_params.update(self._apply_style(i))
            if id(pane) in self._panels:
                card = self._panels[id(pane)]
                card.param.update(**child_params)
            else:
                card = Card(
                    pane,
                    css_classes=['accordion'],
                    header_css_classes=['accordion-header'],
                    **child_params
                )
                card.param.watch(self._set_active, ['collapsed'])
                self._panels[id(pane)] = card
            if ref in card._models:
                panel = card._models[ref][0]
                old_models.append(panel)
            else:
                try:
                    panel = card._get_model(doc, root, model, comm)
                    if self.toggle:
                        cb = CustomJS(args={'accordion': model}, code=self._toggle)
                        panel.js_on_change('collapsed', cb)
                except RerenderError as e:
                    if e.layout is not None and e.layout is not self:
                        raise e
                    e.layout = None
                    return self._get_objects(
                        model, current_objects[:i], doc, root, comm
                    )
            new_models.append(panel)

        self._updating_active = False
        self._set_active()
        self._update_cards()
        self._update_active()
        return new_models, old_models

    def _compute_sizing_mode(self, children, props):
        children = [subchild for child in children for subchild in child.children[1:]]
        return super()._compute_sizing_mode(children, props)

    def _cleanup(self, root: Model | None = None) -> None:
        for panel in self._panels.values():
            panel._cleanup(root)
        super()._cleanup(root)

    def _apply_style(self, i):
        if i == 0:
            margin = (5, 5, 0, 5)
        elif i == (len(self)-1):
            margin = (0, 5, 5, 5)
        else:
            margin = (0, 5, 0, 5)
        return dict(margin=margin, collapsed = i not in self.active)

    def _set_active(self, *events):
        if self._updating_active:
            return
        self._updating_active = True
        try:
            if self.toggle and events and not events[0].new:
                active = [list(self._panels.values()).index(events[0].obj)]
            else:
                active = []
                for i, pane in enumerate(self.objects):
                    if id(pane) not in self._panels:
                        continue
                    elif not self._panels[id(pane)].collapsed:
                        active.append(i)

            if not self.toggle or active:
                self.active = active
        finally:
            self._updating_active = False

    def _update_active(self, *events):
        if self._updating_active:
            return
        self._updating_active = True
        try:
            for i, pane in enumerate(self.objects):
                if id(pane) not in self._panels:
                    continue
                self._panels[id(pane)].collapsed = i not in self.active
        finally:
            self._updating_active = False

    def _update_cards(self, *events):
        params = {
            k: v for k, v in self.param.values().items()
            if k in self._synced_properties
        }
        for panel in self._panels.values():
            panel.param.update(**params)

    # Public API

    def select(
        self, selector: type | Callable[[Viewable], bool] | None = None
    ) -> List[Viewable]:
        selected = Reactive.select(self, selector)
        if self._panels:
            for card in self._panels.values():
                selected += card.select(selector)
            return selected
        return selected
